/***
 *
 * NNPDF2.3QED Standalone Driver v1.0.0
 *
 * Stefano Carrazza for the NNPDF Collaboration
 * email: stefano.carrazza@mi.infn.it
 *
 * August 2013
 *
 * Usage:
 *
 *  NNPDFDriver *pdf = new NNPDFDriver("gridname.LHgrid");
 *
 *  pdf->xfxphoton(x,Q,fl); // -> returns double
 *  // with fl = [-6,7], LHAPDF format
 *    or
 *  pdf->xfxphoton(x,Q); // -> returns double*
 *  // with double* index from [0,13], always in the LHAPDF format
 *
 */

#include "NNPDFDriver.h"
#include <fstream>
#include <stdlib.h>
#include <cmath>

#define XMINGRID 1e-7

using namespace std;

NNPDFDriver::NNPDFDriver(string const& gridfilename):
  fNFL(14),
  fNX(100),
  fNQ2(50),
  fPDFGrid(NULL),
  fXGrid(NULL),
  fQ2Grid(NULL)
{
  // Read PDFs from file
  readPDFSet(gridfilename);
}

NNPDFDriver::~NNPDFDriver()
{
  if (fPDFGrid) {
    for (int i = 0; i < fNFL; i++)
      {
	for (int j = 0; j < fNX; j++)
	  if (fPDFGrid[i][j]) delete[] fPDFGrid[i][j];
	if (fPDFGrid[i]) delete[] fPDFGrid[i];
      }
    delete[] fPDFGrid;
  }

  if (fXGrid) delete[] fXGrid;
  if (fQ2Grid) delete[] fQ2Grid;
}

double* NNPDFDriver::xfxphoton(double const& x, double const& Q)
{
  return xfxevolve(x,Q);
}

double NNPDFDriver::xfxphoton(double const& x, double const& Q, int const& id)
{
  return xfxevolve(x,Q,id+6)[0];
}

double* NNPDFDriver::xfxevolve(double x, double Q, int id)
{
  const double Q2 = Q*Q;

  // check bounds
  if (x < XMINGRID || x > fXGrid[fNX-1]) {
    cout << "Parton interpolation: x out of range -- freezed" << endl;  
    if (x < XMINGRID)  x = XMINGRID;
    if (x > fXGrid[fNX-1]) x = fXGrid[fNX-1];
  }
  if (Q2 < fQ2Grid[0] || Q2 > fQ2Grid[fNQ2-1]) {
    cout << "Parton interpolation: Q2 out of range -- freezed" << endl;  
    if (Q2 < fQ2Grid[0]) x = fQ2Grid[0];
    if (Q2 > fQ2Grid[fNQ2-1]) x = fQ2Grid[fNQ2-1];
  }

  // find nearest points in the grid
  int minx = 0;
  int maxx = fNX;

  while (maxx-minx > 1)
    {
      int midx = (minx+maxx)/2;
      if (x < fXGrid[midx]) 
	maxx = midx;
      else
	minx = midx;      
    }
  int ix = minx;

  int minq = 0;
  int maxq = fNQ2;
  while (maxq-minq > 1)
    {
      int midq = (minq+maxq)/2;
      if (Q2 < fQ2Grid[midq])
	maxq = midq;
      else
	minq = midq;      
    }
  int iq2 = minq;

  // Assign grid for interpolation. M,N -> order of polyN interpolation
  const int M = 4;
  const int N = 2;
  int *ix1a = new int[M];
  int *ix2a = new int[N];
  double *x1a = new double[M];
  double *x2a = new double[N];

  for (int i = 0; i < M; i++)
    {
      if (ix+1 >= M/2 && ix+1 <= (fNX-M/2)) ix1a[i] = ix - M/2 + i+1;
      if (ix+1 < M/2) ix1a[i] = i+1;
      if (ix+1 > (fNX-M/2)) ix1a[i] = (fNX-M) + i+1;
	
      // Check grids
      if (ix1a[i] < 0 || ix1a[i] >= fNX)
	{
	  cout << "Error in grids! i, ixia[i] = " 
	       << i << "\t" << ix1a[i] << endl;
	  exit(-1);
	}
    }

  for (int j = 0; j < N; j++)
    {
      if (iq2+1 >= N/2 && iq2 <= (fNQ2-N/2)) ix2a[j] = iq2 - N/2 + j+1;
      if (iq2+1 < N/2) ix2a[j] = j+1;
      if (iq2+1 > (fNQ2-N/2)) ix2a[j] = (fNQ2-N) + j+1;

      // Check grids
      if (ix2a[j] < 0 || ix2a[j] >= fNQ2)
	{
	  cout << "Error in grids! j, ix2a[j] = "
	       << j << "\t" << ix2a[j] << endl;
	  exit(-1);
	}
    }

  // define points where to evaluate interpolation
  // choose between linear or logarithmic (x,Q2) interpolation
  const double xch = 1e-1;

  double x1;
  if (x < xch) x1 = log(x);
  else x1 = x;
  double x2 = log(Q2);

  double **ya = new double*[M];
  double *xpdf = new double[fNFL];
  for (int i = 0; i < fNFL; i++) xpdf[i] = 0.0;
  
  if (id < 0 && id > 13)
    {
      for (int ipdf = 0; ipdf < fNFL; ipdf++)
	{
	  // Choose betwen linear or logarithmic (x,Q2) interpolation
	  for (int i = 0; i < M; i++)
	    {
	      if (x < xch)
		x1a[i] = log(fXGrid[ix1a[i]]);
	      else
		x1a[i] = fXGrid[ix1a[i]];
	      
	      ya[i] = new double[N];
	      for (int j = 0; j < N; j++)
		{
		  x2a[j] = log(fQ2Grid[ix2a[j]]);
		  ya[i][j] = fPDFGrid[ipdf][ix1a[i]][ix2a[j]];
		}
	    }
	  
	  // 2D polynomial interpolation
	  double y = 0, dy = 0;
	  lh_polin2(x1a,x2a,ya,M,N,x1,x2,y,dy);
	  
	  xpdf[ipdf] = y;
	}
    }
  else
    {
      // Choose betwen linear or logarithmic (x,Q2) interpolation
      for (int i = 0; i < M; i++)
	{
	  if (x < xch)
	    x1a[i] = log(fXGrid[ix1a[i]]);
	  else
	    x1a[i] = fXGrid[ix1a[i]];
	  
	  ya[i] = new double[N];
	  for (int j = 0; j < N; j++)
	    {
	      x2a[j] = log(fQ2Grid[ix2a[j]]);
	      ya[i][j] = fPDFGrid[id][ix1a[i]][ix2a[j]];
	    }
	}
      
      // 2D polynomial interpolation
      double y = 0, dy = 0;
      lh_polin2(x1a,x2a,ya,M,N,x1,x2,y,dy);
      
      xpdf[0] = y;
    }

  // Free memory
  delete[] ix1a;
  delete[] ix2a;
  delete[] x1a;
  delete[] x2a;

  for (int i = 0; i < N; i++)
    delete[] ya[i];
  delete[] ya;

  return xpdf;
}

void NNPDFDriver::readPDFSet(string const& grid)
{
  fstream f;
  f.open(grid.c_str(),ios::in);
  if (f.fail())
    {
      cout << "Error: cannot open file " << grid << endl;
      exit(-1);
    }

  // Reading grid: removing header
  string tmp;
  for (;;)
    {
      getline(f,tmp);
      if (tmp.find("NNPDF20intqed") != string::npos) {
	getline(f,tmp);
	break;	
      }
    }

  // getting nx
  f >> fNX;
  fXGrid = new double[fNX];
  for (int ix = 0; ix < fNX; ix++)
    f >> fXGrid[ix];

  f >> fNQ2;
  f >> tmp; // skip first value
  fQ2Grid = new double[fNQ2];
  for (int iq = 0; iq < fNQ2; iq++)
    f >> fQ2Grid[iq];

  // Prepare grid array  
  fPDFGrid = new double**[fNFL];
  for (int i = 0; i < fNFL; i++)
    {
      fPDFGrid[i] = new double*[fNX];
      for (int j = 0; j < fNX; j++)
	{
	  fPDFGrid[i][j] = new double[fNQ2];
	  for (int z = 0; z < fNQ2; z++)
	    fPDFGrid[i][j][z] = 0.0;
	}
    }  

  // read PDF grid points
  f >> tmp; // remove replica number
  for (int ix = 0; ix < fNX; ix++)
    for (int iq = 0; iq < fNQ2; iq++)
      for (int fl = 0; fl < fNFL; fl++)
	f >> fPDFGrid[fl][ix][iq];

  f.close();
}

void NNPDFDriver::lh_polin2(double *x1a, double *x2a, 
			    double **ya, int m, int n, 
			    double x1, double x2, 
			    double& y, double& dy)
{
  double *yntmp = new double[n];
  double *ymtmp = new double[m];

  for (int j = 0; j < m; j++)
    {
      for(int k = 0; k < n; k++)
	yntmp[k] = ya[j][k];
      
      lh_polint(x2a,yntmp,n,x2,ymtmp[j],dy);
    }
  lh_polint(x1a,ymtmp,m,x1,y,dy);
  
  delete[] yntmp;
  delete[] ymtmp;
}

void NNPDFDriver::lh_polint(double *xa, double *ya, int n, double x, 
			    double& y, double& dy)
{
  int ns = 0;  
  double dif = abs(x-xa[0]);
  double *c = new double[n];
  double *d = new double[n];
  
  for (int i = 0; i < n; i++)
    {
      double dift = abs(x-xa[i]);
      if (dift < dif)
	{
	  ns = i;
	  dif = dift;
	}
      c[i] = ya[i];
      d[i] = ya[i];
    }
  y = ya[ns];
  ns--;
  for (int m = 1; m < n; m++)
    {
      for (int i = 0; i < n-m; i++)
	{
	  double ho = xa[i]-x;
	  double hp = xa[i+m]-x;
	  double w = c[i+1]-d[i];
	  double den = ho-hp;
	  if (den == 0)
	    {
	      cout << "failure in polint" << endl;
	      exit(-1);	       
	    }
	  den = w/den;
	  d[i] = hp*den;
	  c[i] = ho*den;
	}
      if (2*(ns+1) < n-m)
	dy = c[ns+1];
      else {
	dy = d[ns];
	ns--;
      }
      y+=dy;
    }

  delete[] c;
  delete[] d;
}
